package eu.jonahbauer.raytracing.render;

import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.shape.Shape;
import org.jetbrains.annotations.NotNull;

import java.util.List;

public record Scene(@NotNull List<@NotNull Shape> shapes) {

    public Scene {
        shapes = List.copyOf(shapes);
    }

    public Scene(@NotNull Shape @NotNull ... shapes) {
        this(List.of(shapes));
    }

    public @NotNull Image render(@NotNull Camera camera) {
        var image = new Image(camera.width(), camera.height());

        camera.pixels().forEach(pixel -> {
            var x = pixel.x();
            var y = pixel.y();
            var ray = pixel.ray();

            var result = shapes.stream()
                    .mapToDouble(shape -> shape.hit(ray))
                    .filter(Double::isFinite)
                    .min();
            if (result.isPresent()) {
                image.set(x, y, Color.RED);
            } else {
                image.set(x, y, getSkyboxColor(ray));
            }
        });

        return image;
    }

    private @NotNull Color getSkyboxColor(@NotNull Ray ray) {
        // altitude from -pi/2 to pi/2
        var alt = Math.copySign(
                Math.acos(ray.direction().withY(0).unit().times(ray.direction().unit())),
                ray.direction().y()
        );
        return Color.lerp(Color.WHITE, Color.SKY, alt / Math.PI + 0.5);
    }
}
